#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test on LDAP service object """

import pytest

from eesyldap.exceptions import LDAPRelatedObjectModifed
from eesyldap.exceptions import LDAPRequiredAttribute
from eesyldap.search import LDAPSearch


def test_service_attributes_access(dev):
    """ Test access to service's attributes """
    assert dev.id == 'dev'
    assert dev.name == 'Dev'
    assert dev.mail == 'dev@easter-eggs.com'
    assert dev.numid == 3
    assert dev.manager_dn == 'uid=brenard,ou=people,o=eesyldap'


def test_service_relations_access(dev, brenard, ee, Group):     # pylint: disable=unused-argument
    """
    Test access to company relations

    Note : Group parameter is need to be sure Group object type is declared.
    """

    # Groups
    groups = dev.groups
    assert isinstance(groups, LDAPSearch) and groups.filterstr == '(&(objectClass=top)(objectClass=elGroup)(ou=dev))'
    assert not groups

    # Manager
    assert dev.manager == brenard

    # Users
    users = dev.users
    assert isinstance(users, LDAPSearch) and users.filterstr == '(&(objectClass=top)(objectClass=inetOrgPerson)(objectClass=elPeople)(ou=dev))'
    assert len(users) == 1
    assert brenard.username in users
    assert users[brenard.username] == brenard

    # Company
    assert dev.company == ee


def test_service_remove_company(dev):
    """ Test to remove service's company """
    with pytest.raises(LDAPRequiredAttribute):
        dev.company = None


def test_service_add_user(dev, brenard):
    """ Test to add a user to service """
    users = dict(dev.users)
    users[brenard.username] = brenard
    dev.users = users

    # Verify
    brenard.refresh()
    assert 'dev' in brenard.service_ids


def test_service_remove_user(admin, brenard):
    """ Test to remove a user to service """
    users = dict(admin.users)
    del users[brenard.username]
    admin.users = users

    # Verify
    brenard.refresh()
    assert 'admin' not in brenard.service_ids


def test_service_add_modified_user(dev, jdoe):
    """ Test to add a modified user to service """
    jdoe.numid = 88
    users = dict(dev.users)
    users[jdoe.username] = jdoe

    with pytest.raises(LDAPRelatedObjectModifed):
        dev.users = users


def test_service_set_manager(dev, jdoe):
    """ Test to set manager of a service """
    dev.manager = jdoe
    dev.refresh()

    assert dev.manager_dn == jdoe.dn


def test_service_set_modified_manager(dev, brenard):
    """ Test to set a modified user as manager of a service """
    brenard.numid = 77
    with pytest.raises(LDAPRelatedObjectModifed):
        dev.manager = brenard


def test_service_remove_company_using_reverse_relation(dev, ee):
    """ Test to remove a service's company using the reverse relation """
    services = dict(ee.services)
    del services[dev.id]
    with pytest.raises(LDAPRequiredAttribute):
        ee.services = services



# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test LDAP Password attribute type """

import pytest
from unittest.mock import patch

from eesyldap.attributes import Password
from eesyldap.object import LDAPObject

# pylint: disable=protected-access,redefined-outer-name


def test_create_all_hash_types():
    """ Test to create LDAP Password attribute with each supported hash types """
    for hash_type in Password._hash_types:
        Password('fakeAttr', hash_type=hash_type)


def test_create_without_hash_types():
    """ Test to create LDAP Password attribute without specify hash type """
    Password('fakeAttr')


def test_create_bad_hash_type():
    """ Test to create LDAP Password attribute with bad hash type """
    with pytest.raises(AssertionError):
        Password('fakeAttr', hash_type='fake')


def test_hash_and_verify_md5():
    """ Test to hash and verify a MD5 password """
    attr = Password('fakeAttr', hash_type='md5')
    clear_password = 'azerty'
    hashed_password = attr.hash(clear_password)

    assert hashed_password == '{MD5}q09j+axlFSV1iGhg3eSAoQ=='

    assert attr.verify(clear_password, hashed_passwords=[hashed_password])


def test_hash_and_verify_md5_crypt():
    """ Test to hash and verify a MD5 CRIPT password """
    attr = Password('fakeAttr', hash_type='md5_crypt')
    clear_password = 'azerty'
    hashed_password = attr.hash(clear_password)

    assert isinstance(hashed_password, str) and hashed_password.startswith('{CRYPT}$1$')
    assert len(hashed_password) == 41

    assert attr.verify(clear_password, hashed_passwords=[hashed_password])


def test_hash_and_verify_sha():
    """ Test to hash and verify a SHA password """
    attr = Password('fakeAttr', hash_type='sha')
    clear_password = 'azerty'
    hashed_password = attr.hash(clear_password)

    assert hashed_password == '{SHA}nPldrNIm3PQ9o3bNtsu6cDUhiSE='

    assert attr.verify(clear_password, hashed_passwords=[hashed_password])


def test_hash_and_verify_ssha():
    """ Test to hash and verify a SSHA password """
    attr = Password('fakeAttr', hash_type='ssha')
    clear_password = 'azerty'
    hashed_password = attr.hash(clear_password)

    assert isinstance(hashed_password, str) and hashed_password.startswith('{SSHA}')
    assert len(hashed_password) == 38

    assert attr.verify(clear_password, hashed_passwords=[hashed_password])


def test_hash_and_verify_ssha256():
    """ Test to hash and verify a SSHA256 password """
    attr = Password('fakeAttr', hash_type='ssha256')
    clear_password = 'azerty'
    hashed_password = attr.hash(clear_password)

    assert isinstance(hashed_password, str) and hashed_password.startswith('{CRYPT}$5$')
    assert len(hashed_password) == 84

    assert attr.verify(clear_password, hashed_passwords=[hashed_password])


def test_hash_and_verify_ssha512():
    """ Test to hash and verify a SSHA512 password """
    attr = Password('fakeAttr', hash_type='ssha512')
    clear_password = 'azerty'
    hashed_password = attr.hash(clear_password)

    assert isinstance(hashed_password, str) and hashed_password.startswith('{CRYPT}$6$')
    assert len(hashed_password) == 127

    assert attr.verify(clear_password, hashed_passwords=[hashed_password])


def test_hash_and_verify_clear():
    """ Test to hash and verify a clear password """
    attr = Password('fakeAttr', hash_type='clear')
    clear_password = 'azerty'
    hashed_password = attr.hash(clear_password)

    assert hashed_password == clear_password

    assert attr.verify(clear_password, hashed_passwords=[hashed_password])


def test_verify_clear_password():
    """ Test to verify a clear password """
    attr = Password('fakeAttr', hash_type='ssha256')
    assert attr.verify('clear', hashed_passwords=['clear'])
    assert not attr.verify('clear', hashed_passwords=['not_clear'])


def test_verify_without_hashed_passwords(brenard):
    """
    Test to verify password without specify hashed_passwords but providing LDAP object """
    assert brenard._attrs['password'].verify('password', obj=brenard)


def test_is_hashed_password_when_true():
    """ Test the is_hashed_password method when it is True """
    assert Password.is_hashed_password('{SHA}nPldrNIm3PQ9o3bNtsu6cDUhiSE=')


def test_is_hashed_password_when_false():
    """ Test the is_hashed_password method when it is not True """
    assert not Password.is_hashed_password('clear')


def test_to_ldap_clear_password():
    """ Test LDAP Password attribute _to_ldap method with a clear password """
    attr = Password('fakeAttr', hash_type='sha')
    assert attr._to_ldap('ascii') == b'{SHA}UcBms26osyB2lkx2b4oDJMpOtLk='


def test_to_ldap_hashed_password():
    """ Test LDAP Password attribute _to_ldap method with a hashed password """
    attr = Password('fakeAttr', hash_type='sha')
    assert attr._to_ldap('{SHA}qUqP5cyxm6YcTAhz05Hph5gvu9M=') == b'{SHA}qUqP5cyxm6YcTAhz05Hph5gvu9M='



# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test LDAP Datetime attribute type """

from datetime import datetime
import pytest
import pytz

from eesyldap.attributes import Datetime

# pylint: disable=protected-access,redefined-outer-name


@pytest.fixture
def tz_paris():
    """ Retreive Europe/Paris timezone object """
    return pytz.timezone('Europe/Paris')


def test_create_with_timezone_local():
    """ Test to create LDAP Datetime attribute with timezone local """
    Datetime('fakeAttr', python_timezone='local', ldap_timezone='local')


def test_create_with_timezone_europe_paris():
    """ Test to create LDAP Datetime attribute with timezone Europe/Paris """
    Datetime('fakeAttr', python_timezone='Europe/Paris', ldap_timezone='Europe/Paris')


def test_create_with_timezone_object(tz_paris):
    """ Test to create LDAP Datetime attribute with a timezone object """
    Datetime('fakeAttr', python_timezone=tz_paris, ldap_timezone=pytz.utc)


def test_create_with_naive_timezone():
    """ Test to create LDAP Datetime attribute with naive timezone """
    Datetime('fakeAttr', naive=True)


@pytest.fixture
def fake_attr_naive():
    """ Generate a fake LDAP Date attribute with naive timezone """
    return Datetime('fakeAttr', naive=True)


def test_to_python_naive(fake_attr_naive):
    """ Test LDAP Datetime attribute _to_python method with naive timezone"""
    assert fake_attr_naive._to_python(b'19971010000000Z') == datetime(1997, 10, 10, 0, 0, 0)
    assert fake_attr_naive._to_python(b'20200116121212Z') == datetime(2020, 1, 16, 12, 12, 12)
    assert fake_attr_naive._to_python(b'20200116121212+0100') == datetime(2020, 1, 16, 12, 12, 12)
    assert fake_attr_naive._to_python(None) is None


def test_to_python_sortable_naive(fake_attr_naive):
    """ Test LDAP Datetime attribute _to_python method with sortable=True and naive timezone """
    assert fake_attr_naive._to_python(b'19971010000000Z', sortable=True) == '19971010000000'
    assert fake_attr_naive._to_python(b'20200116121212Z', sortable=True) == '20200116121212'
    assert fake_attr_naive._to_python(b'20200116121212+0100', sortable=True) == '20200116121212'
    assert fake_attr_naive._to_python(None, sortable=True) == ''


def test_to_ldap_naive(fake_attr_naive):
    """ Test LDAP Datetime attribute _to_ldap method with naive timezone """
    assert fake_attr_naive._to_ldap(datetime(1997, 11, 21, 15, 12, 54)) == b'19971121151254Z'
    assert fake_attr_naive._to_ldap(datetime(2020, 1, 17, 00, 23, 55)) == b'20200117002355Z'
    with pytest.raises(AssertionError):
        fake_attr_naive._to_ldap(None)


@pytest.fixture
def fake_attr_utc_paris(tz_paris):
    """
    Generate a fake LDAP Date attribute with UTC timezone in LDAP
    and Europe/Paris timezone on Python side.
    """
    return Datetime('fakeAttr', ldap_timezone=pytz.utc, python_timezone=tz_paris)


def test_to_python_utc_paris(fake_attr_utc_paris, tz_paris):
    """ Test LDAP Datetime attribute _to_python method with naive timezone"""
    assert fake_attr_utc_paris._to_python(b'19971010000000Z') == datetime(1997, 10, 10, 0, 0, 0, tzinfo=pytz.utc).astimezone(tz_paris)
    assert fake_attr_utc_paris._to_python(b'20200116121212Z') == datetime(2020, 1, 16, 12, 12, 12, tzinfo=pytz.utc).astimezone(tz_paris)
    assert fake_attr_utc_paris._to_python(b'20200116121212+0100') == datetime(2020, 1, 16, 11, 12, 12, tzinfo=pytz.utc).astimezone(tz_paris)
    assert fake_attr_utc_paris._to_python(None) is None


def test_to_ldap_utc_paris(fake_attr_utc_paris, tz_paris):
    """ Test LDAP Datetime attribute _to_python method with naive timezone"""
    assert fake_attr_utc_paris._to_ldap(datetime(1997, 11, 21, 15, 12, 54, tzinfo=pytz.utc).astimezone(tz_paris)) == b'19971121151254Z'
    assert fake_attr_utc_paris._to_ldap(datetime(2020, 1, 17, 00, 23, 55, tzinfo=pytz.utc).astimezone(tz_paris)) == b'20200117002355Z'
    with pytest.raises(AssertionError):
        fake_attr_utc_paris._to_ldap(None)


@pytest.fixture
def fake_attr_paris_paris(tz_paris):
    """
    Generate a fake LDAP Date attribute with Europe/Paris timezone in both
    LDAP and Python sides.
    """
    return Datetime('fakeAttr', ldap_timezone=tz_paris, python_timezone=tz_paris)


def test_to_python_paris_paris(fake_attr_paris_paris, tz_paris):
    """ Test LDAP Datetime attribute _to_python method with naive timezone"""
    assert fake_attr_paris_paris._to_python(b'19971010000000+0200') == datetime(1997, 10, 9, 22, 0, 0, tzinfo=pytz.utc).astimezone(tz_paris)
    assert fake_attr_paris_paris._to_python(b'20200716121212+0100') == datetime(2020, 7, 16, 11, 12, 12, tzinfo=pytz.utc).astimezone(tz_paris)
    assert fake_attr_paris_paris._to_python(b'20200116121212Z') == datetime(2020, 1, 16, 12, 12, 12, tzinfo=pytz.utc).astimezone(tz_paris)
    assert fake_attr_paris_paris._to_python(None) is None


def test_to_ldap_paris_paris(fake_attr_paris_paris, tz_paris):
    """ Test LDAP Datetime attribute _to_python method with naive timezone"""
    assert fake_attr_paris_paris._to_ldap(datetime(1997, 11, 21, 15, 12, 54, tzinfo=pytz.utc).astimezone(tz_paris)) == b'19971121161254+0100'
    assert fake_attr_paris_paris._to_ldap(datetime(2020, 7, 17, 00, 23, 55, tzinfo=pytz.utc).astimezone(tz_paris)) == b'20200717022355+0200'
    with pytest.raises(AssertionError):
        fake_attr_paris_paris._to_ldap(None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP Client """

from collections import OrderedDict
import logging

from beaker.cache import Cache
from beaker.util import parse_cache_config_options
import ldap.modlist
from ldap import LDAPError  # pylint: disable=no-name-in-module

from eesyldap.connection import LDAPConnection
from eesyldap.exceptions import LDAPDuplicateObjectFound
from eesyldap.exceptions import LDAPInvalidAttributeName
from eesyldap.exceptions import LDAPNotUniqueAttribute
from eesyldap.exceptions import LDAPSearchLimitExceeded
from eesyldap.helpers import format_modlist
from eesyldap.object import LDAPObject
from eesyldap.search import LDAPSearch

log = logging.getLogger(__name__)


class LDAPClient:
    """
    LDAP client

    This class abstract all interactions with this LDAP server.
    """

    def __init__(self, config, **kwargs):
        """
        LDAP Client constructor

        :param  config:     LDAP client configuration

        All keyword arguments starting by 'cache_' will be used to configure Beaker object cache.
        See Beaker documentation for more details : https://beaker.readthedocs.io/en/latest/configuration.html#configuration
        """
        self.config = config

        # Cache
        self.__cache = {}
        self.cache_params = parse_cache_config_options(dict(
            ('cache.{0}'.format(k[6:]), v)
            for k, v in kwargs.items()
            if k.startswith('cache_')
        ))

        self.con = LDAPConnection(config)

    #
    # Cache helpers
    #
    def _get_cache(self, cache_namespace):
        """ Retreive/initialize cache (if enabled) for a specific namespace """
        if self.cache_params.get('enabled'):
            try:
                return self.__cache[cache_namespace]
            except KeyError:
                self.__cache[cache_namespace] = cache = Cache(cache_namespace, **self.cache_params)
                log.info('[LDAP] Cache %s enabled, type=%s, expire=%s.', cache_namespace, self.cache_params.get('type'), self.cache_params.get('expire'))
                return cache
        else:
            log.debug('[LDAP] Cache is disabled.')
        return None

    #
    # Object cache
    #
    # This cache permit to store object's data.
    #
    def _get_object_cache(self, obj_type):
        """ Return object type cache namespace """
        return self._get_cache("ldap.objects.%s" % obj_type.__name__)

    @staticmethod
    def get_cached_object_identifiers(obj):
        """
        Retreive all object identifier based on all unique attribute attr_values

        :param obj:             The object
        """

        identifiers = {'dn': obj.dn}
        for attr_name, attr_value in obj.items():
            if obj.is_unique(attr_name) and attr_value:
                identifiers[attr_name] = str(attr_value)
        return identifiers

    @staticmethod
    def get_object_cache_key(identifier, value):
        """
        Retrieve objects cache key

        :param identifier:     The LDAP object identifier attribute name
        :param value:          The LDAP object identifier attribute value
        """
        return '{0}={1}'.format(identifier, value)

    def add_cached_object(self, obj_type, obj):
        """
        Add on object in cache (if enabled)

        :param obj_type:       The object type
        :param obj:            The object
        """
        cache = self._get_object_cache(obj_type)
        if not cache:
            return

        identifiers = self.get_cached_object_identifiers(obj)

        serialized_obj = obj.serialize()
        for identifier, value in identifiers.items():
            if value:
                cache_key = self.get_object_cache_key(identifier, value)
                cache[cache_key] = serialized_obj
                log.debug('Object %s added to cache %s with key "%s" : %s', obj, cache, cache_key, serialized_obj)

    def get_cached_object(self, obj_type, identifiers=None, refresh=False):
        """
        Retrieve on object from cache (if enabled)

        :param obj_type:       The object type
        :param identifiers:    A dict of the object identifier(s) :
                                - key : the attribute name
                                - value : the value of this attribute that refer only
                                to this specific object.
        :param refresh:        Helper : force cache update if True
        """
        cache = self._get_object_cache(obj_type)
        if cache and not refresh:
            for identifier, value in identifiers.items():
                if not value:
                    continue
                cache_key = self.get_object_cache_key(identifier, value)
                try:
                    serialized_obj = cache[cache_key]
                    obj = obj_type.unserialize(self, serialized_obj)
                    log.debug('%s object %s retreived from cache %s (from key = "%s")', obj_type.__name__, obj, cache, cache_key)
                    return obj
                except KeyError:
                    pass
            log.debug('No object %s found in cache %s with identifiers %s', obj_type.__name__, cache, ' / '.join(['%s=%s' % (attr_name, attr_value) for attr_name, attr_value in identifiers.items()]))
        return None

    def delete_cached_object(self, obj_type, obj=None, identifiers=None):
        """
        Delete on object in cache (if enabled and if object is present)

        :param obj_type:       The object type
        :param obj:            The object to delete in cache (to retreive it's identifiers, see get_cached_object())
        :param identifiers:    A dict of the object identifier(s) (see get_cached_object())
        """
        cache = self._get_object_cache(obj_type)
        if not cache:
            return

        if obj is not None:
            identifiers = self.get_cached_object_identifiers(obj)

        for identifier, value in identifiers.items():
            if not value:
                continue
            cache_key = self.get_object_cache_key(identifier, value)
            try:
                del cache[cache_key]
                log.debug('Object %s removed in cache %s with key "%s"', obj, cache, cache_key)
            except KeyError:
                log.debug('Object %s not found in cache %s with key "%s"', obj, cache, cache_key)

    #
    # Object search cache
    #
    # This cache permit to store search result data as retreived from LDAPConnection.search().
    #
    def _get_object_search_cache(self, obj_type):
        """ Return search of object type cache namespace """
        return self._get_cache("ldap.objects.search.%s" % obj_type.__name__)

    @staticmethod
    def get_object_search_cache_key(filterstr, base_dn, scope):
        """
        Retrieve objects search cache key

        :param filterstr:      The LDAP search filter
        :param base_dn:        The LDAP search base DN
        :param scope:          The LDAP search scope
        """
        return 'filter={0}|base_dn={1}|scope={2}'.format(filterstr, base_dn, scope)

    def add_cached_object_search(self, obj_type, filterstr, base_dn, scope, search_result):
        """
        Add objects search in cache (if enabled)

        :param obj_type:       The object type
        :param filterstr:      The LDAP search filter
        :param base_dn:        The LDAP search base DN
        :param scope:          The LDAP search scope
        :param search_result:  The LDAP search result
        """
        cache = self._get_object_search_cache(obj_type)
        if cache:
            cache_key = self.get_object_search_cache_key(filterstr, base_dn, scope)
            cache[cache_key] = search_result
            log.debug('Object %s search added to cache %s with key "%s" (%s object(s))', obj_type.__name__, cache, cache_key, len(search_result))

    def get_cached_object_search(self, obj_type, filterstr, base_dn, scope, refresh=False):
        """
        Retrieve objects search from cache (if enabled)

        :param obj_type:       The object type
        :param filterstr:      The LDAP search filter
        :param base_dn:        The LDAP search base DN
        :param scope:          The LDAP search scope
        :param refresh:        Helper : force cache update if True
        """
        cache = self._get_object_search_cache(obj_type)
        if cache and not refresh:
            cache_key = self.get_object_search_cache_key(filterstr, base_dn, scope)
            try:
                result = cache[cache_key]
                log.debug('Search result of objects %s retreived from cache %s (with key "%s")', obj_type.__name__, cache, cache_key)
                return result
            except KeyError:
                pass
            log.debug('Search result of objects %s not found in cache %s (with key "%s")', obj_type.__name__, cache, cache_key)
        return None

    def delete_cached_object_search(self, obj_type):
        """
        Delete all objects search of a specific type in cache (if cache is enabled)

        :param obj_type:       The object type
        """
        cache = self._get_object_search_cache(obj_type)
        if cache:
            cache.clear()
            log.debug('Object %s search cache %s cleared', obj_type.__name__, cache)

        # Also delete sorted search results
        self.delete_cached_object_sorted_search(obj_type)

    #
    # Object sorted search cache
    #
    # This cache permit to store sorted search result as a list of sorted objects's DN.
    #
    def _get_object_sorted_search_cache(self, obj_type):
        """ Return sorted search of object type cache namespace """
        return self._get_cache("ldap.objects.search.%s.sorted" % obj_type.__name__)

    @staticmethod
    def get_object_sorted_search_cache_key(filterstr, base_dn, scope, sort_by, sort_reverse):
        """
        Retrieve objects sorted search cache key

        :param filterstr:      The LDAP search filter
        :param base_dn:        The LDAP search base DN
        :param scope:          The LDAP search scope
        :param sort_by:        The sort by parameter
        :param sort_reverse:   The sort reverse parameter
        """
        return 'filter={0}|base_dn={1}|scope={2}|sort_by={3}|sort_reverse={4}'.format(filterstr, base_dn, scope, sort_by, bool(sort_reverse))

    def add_cached_object_sorted_search(self, obj_type, filterstr, base_dn, scope, sort_by, sort_reverse, obj_dns):
        """
        Add objects sorted search result in cache (if enabled)

        :param obj_type:       The object type
        :param filterstr:      The LDAP search filter
        :param base_dn:        The LDAP search base DN
        :param scope:          The LDAP search scope
        :param obj_dns:        The objects's DN list
        """
        cache = self._get_object_sorted_search_cache(obj_type)
        if cache:
            cache_key = self.get_object_sorted_search_cache_key(filterstr, base_dn, scope, sort_by, sort_reverse)
            cache[cache_key] = obj_dns
            log.debug('Object %s sorted search result added to cache %s with key "%s" (%s object(s))', obj_type.__name__, cache, cache_key, len(obj_dns))

    def get_cached_object_sorted_search(self, obj_type, filterstr, base_dn, scope, sort_by, sort_reverse, refresh=False):
        """
        Retrieve objects sorted search result from cache (if enabled)

        :param obj_type:       The object type
        :param filterstr:      The LDAP search filter
        :param base_dn:        The LDAP search base DN
        :param scope:          The LDAP search scope
        :param sort_by:        The sort by parameter
        :param sort_reverse:   The sort reverse parameter
        :param refresh:        Helper : force cache update if True
        """
        cache = self._get_object_sorted_search_cache(obj_type)
        if cache and not refresh:
            cache_key = self.get_object_sorted_search_cache_key(filterstr, base_dn, scope, sort_by, sort_reverse)
            try:
                result = cache[cache_key]
                log.debug('Sorted search result of objects %s retreived from cache %s (with key "%s")', obj_type.__name__, cache, cache_key)
                return result
            except KeyError:
                pass
            log.debug('Sorted search result of objects %s not found in cache %s (with key "%s")', obj_type.__name__, cache, cache_key)
        return None

    def delete_cached_object_sorted_search(self, obj_type):
        """
        Delete all objects sorted search of a specific type in cache (if cache is enabled)

        :param obj_type:       The object type
        """
        cache = self._get_object_sorted_search_cache(obj_type)
        if cache:
            cache.clear()
            log.debug('Object %s sorted search cache %s cleared', obj_type.__name__, cache)

    #
    # Config helpers
    #
    def get_config(self, var_name, obj_type=None, default=None):
        """
        Retreive configure variable value

        :param  var_name:   The configuration variable name
        :param  obj_type:   The object type (optional). If defined, we first try to retreive the variable value
                            by prefixing it name by the object type name (in lowercase).
                            Example : for variable 'base_dn' of object type User, we first try to retreive
                            the variable 'user.base_dn'.
        :param  default:    The default value to return if variable is not defined (optinal, default: None)
        """
        if obj_type:
            return self.config.get('%s.%s' % (obj_type.__name__.lower(), var_name), self.get_config(var_name, default=default))
        return self.config.get(var_name, default)

    #
    # Objects access helpers
    #

    def get_object_base_dn(self, obj_type):
        """ Return LDAP search base DN for a specific object type """
        return self.get_config('base_dn', obj_type=obj_type)

    def get_object_scope(self, obj_type):
        """ Return LDAP search scope for a specific object type """
        return self.get_config('scope', obj_type=obj_type, default='sub')

    def get(self, obj_type, refresh=False, **identifiers):
        """
        Get an object in LDAP

        :param  obj_type:       The object type to search
        :param  refresh:        Set to True to be sure that result is not retreive from cache
        :param  identifiers:    Dict of object identifiers
        """

        # Check if object is already in cache (if enabled and not refresh mode)
        cached_object = self.get_cached_object(obj_type, identifiers=identifiers, refresh=refresh)
        if cached_object:
            return cached_object

        # Build object search parameters
        filterstr = None
        base_dn = self.get_object_base_dn(obj_type)
        scope = self.get_object_scope(obj_type)
        if identifiers.get('dn'):
            filterstr = obj_type.get_filter()
            base_dn = identifiers.get('dn')
            scope = 'base'
        elif identifiers:
            # Check identifiers
            for attr_name, attr_value in identifiers.items():
                if obj_type.exists(attr_name):
                    # Check attribute is unique
                    if not obj_type.is_unique(attr_name):
                        raise LDAPNotUniqueAttribute(attr_name)
                    filterstr = obj_type.get_attribute_filter(attr_name, attr_value)
                    break
                if attr_name != 'dn':
                    raise LDAPInvalidAttributeName(attr_name)
        if filterstr is None:
            raise Exception('No identifier specify')

        # Search matching object in LDAP
        try:
            search = self.search(
                obj_type,
                base_dn=base_dn,
                filterstr=filterstr,
                scope=scope,
                limit=1,
                refresh=refresh
            )
            if search:
                obj = search[0]
            else:
                log.debug('No matching %s object found.', obj_type.__name__)
                return None
        except LDAPSearchLimitExceeded:
            raise LDAPDuplicateObjectFound(obj_type, filterstr, base_dn, scope)

        if isinstance(obj, LDAPObject):
            log.debug('One corresponding %s object found in LDAP : %s', obj_type.__name__, obj)
            self.add_cached_object(obj_type, obj)
            return obj

        return None

    def raw_search(self, filterstr, base_dn=None, scope=None, attrlist=None, limit=None, hardlimit=None):
        """
        Run raw search using LDAP Connection and return OrderedDict with object DN as key and a dict of LDAP attributes
        values as correponding value.

        :param  filterstr:      The LDAP filter string
        :param  base_dn:        The base DN of the search (optional, default = directory base DN)
        :param  scope:          The scope of the search (optional, default: see LDAPConnection._get_scope())
        :param  attrlist:       The list of requested attributes (optional, default: all)
        :param  limit:          Specific the maximum number of object you want to retreived (optional, default: no limit)
                                If the limit is exceeded, an LDAPSearchLimitExceeded exception will be raised.
        :param  hardlimit:      Specific the maximum number of object you want to retreived (optional, default: no limit)
                                This limit is handled by LDAP server and potentially included references. Prefer limit
                                parameter usage to avoid counting references. If the limit is exceeded, an
                                LDAPSearchHardLimitExceeded exception will be raised.
        """
        base_dn = self.get_config('base_dn') if base_dn is None else base_dn
        with self.con:
            try:
                result = self.con.search(
                    base=base_dn,
                    filterstr=filterstr,
                    scope=scope,
                    attrlist=attrlist,
                    limit=limit,
                    hardlimit=hardlimit
                )
            except LDAPSearchLimitExceeded as err:
                raise err
            except Exception:
                log.error('Error occured searching with filter %s on %s (scope = %s)', filterstr, base_dn, scope, exc_info=1)
                return False

        if not result:
            log.debug('No object found in LDAP with filter %s on %s (scope = %s, attrs = %s)', filterstr, base_dn, scope, attrlist)
            return OrderedDict()

        return OrderedDict(
            (obj_dn, obj_attrs_values)
            for obj_dn, obj_attrs_values in result
        )

    def search(self, obj_type, **kwargs):
        """
        Search objects in LDAP

        This method return an LDAPSearch object.
        """
        return LDAPSearch(self, obj_type, **kwargs)

    def new(self, obj_type, attrs=None, save=False):
        """
        Create new object

        :param  obj_type:       The object type to create
        :param  attrs:          The attributes values to set on new obj (optional)
        :param  save:           Directly save or not this new obj on LDAP directory (only if attrs is provided)
        """
        obj = obj_type(self)
        if isinstance(attrs, dict):
            obj.update(attrs, save=save)
        return obj

    def modify(self, obj, old_attrs, new_attrs, force_replace_attrs=None):
        """
        Modify an object in LDAP directory

        :param  obj:                    The modified object
        :param  old_attrs:              Dict of old LDAP attributes values
        :param  new_attrs:              Dict of the LDAP attributes values
        :param  force_replace_attrs:    List LDAP attributes for which we must use the replace operation on change
        """
        # Generate ignored attributes list
        ignored_attrs = obj.protected_attrs if obj.protected_attrs else []

        # Ignore all force replace attributes (if present in new attributes and not empty)
        if force_replace_attrs:
            for attr in force_replace_attrs:
                if attr in new_attrs and new_attrs[attr]:
                    ignored_attrs.append(attr)

        # Generate modlist
        modlist = ldap.modlist.modifyModlist(
            old_attrs, new_attrs,
            ignore_attr_types=ignored_attrs
        )

        # Add replace operations for force replace attributes (if present in new attributes and not empty)
        if force_replace_attrs:
            for attr in force_replace_attrs:
                if attr in new_attrs and new_attrs[attr]:
                    modlist.append(
                        (ldap.MOD_REPLACE, attr, new_attrs[attr])  # pylint: disable=no-member
                    )

        if not modlist:
            log.debug('No change detected on object %s', obj.dn)
            return True
        log.debug(
            'LDAP changes detected on object %s :\n%s',
            obj.dn, format_modlist(modlist, prefix='  ')
        )

        # Update object in LDAP
        with self.con:
            try:
                self.con.modify(obj.dn, modlist)
                log.debug('Object %s updated in LDAP', obj.dn)
                self.delete_cached_object(obj.__class__, obj)
                self.delete_cached_object_search(obj.__class__)
                return True
            except LDAPError:
                log.error("Error updating %s : %s", obj.dn, modlist, exc_info=1)
        return False

    def rename(self, obj, new_rdn):
        """
        Modify an object in LDAP directory

        :param  obj:            The modified object
        :param  new_rdn:        The LDAP object new RDN (ex : "uid=jdoe")
        """

        # Rename object in LDAP
        with self.con:
            try:
                self.con.rename(obj.dn, new_rdn)
                log.debug('Object %s renamed in LDAP as %s', obj.dn, new_rdn)
                self.delete_cached_object(obj.__class__, obj)
                self.delete_cached_object_search(obj.__class__)
                return True
            except LDAPError:
                log.error("Error renaming %s as %s", obj.dn, new_rdn, exc_info=1)
        return False

    def add(self, obj, dn, attrs):
        """
        Add an object in LDAP directory

        :param  obj:    The object to add
        """
        with self.con:
            try:
                modlist = ldap.modlist.addModlist(attrs)
                if self.con.add(dn, modlist):
                    log.debug('Object %s added in LDAP', dn)
                    self.delete_cached_object_search(obj.__class__)
                    return True
            except LDAPError:
                log.error("Error adding %s : %s", dn, modlist, exc_info=1)
        return False

    def delete(self, obj):
        """ Delete an object in LDAP directory """
        with self.con:
            try:
                dn = obj.dn
                if dn is None:
                    log.error('Fail to retreive object DN')
                    return False
                if self.con.delete(dn):
                    log.debug('Object %s deleted in LDAP', dn)
                    self.delete_cached_object(obj.__class__, obj)
                    self.delete_cached_object_search(obj.__class__)
                    return True
            except LDAPError:
                log.error("Error deleting %s", dn, exc_info=1)
        return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eesyldap', 'tests']

package_data = \
{'': ['*'],
 'tests': ['attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py']}

install_requires = \
['beaker>=1.11.0,<2.0.0',
 'passlib>=1.7.2,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-ldap',
 'pytz>=2019.3,<2020.0',
 'unidecode>=1.1.1,<2.0.0']

setup_kwargs = {
    'name': 'eesyldap',
    'version': '0.3.0',
    'description': 'Simple LDAP Objects mapper',
    'long_description': '# EesyLDAP\n\nThis library offer an easy way to manipulate LDAP objects.\n\n## Getting Started\n\n    git clone https://gitlab.easter-eggs.com/brenard/eesyldap.git\n    cd eesyldap\n    python3 -m venv venv\n    source venv/bin/activate\n    pip install poetry\n    poetry install\n\nSee ```doc/quickstart.md``` to learn how to use this library.\n\n## Run tests\n\nTo run tests, you need to install eesyldap (see _Getting Started_ section) and run the following command being in the root folder of the project : `pytest --disable-warnings --verbose`\n\n## License\nEesyLDAP is free software; you can redistribute it and/or modify it under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.\n\nEesyLDAP is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.\n\nYou should have received a copy of the GNU Affero General Public License along with this program.  If not, see <http://www.gnu.org/licenses/>.\n',
    'author': 'Benjamin Renard',
    'author_email': 'brenard@easter-eggs.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.easter-eggs.com/brenard/eesyldap',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test on LDAP group object """


def test_group_attributes_access(ldapsuperadmins):
    """ Test access to group's attributes """
    assert ldapsuperadmins.name == 'ldapsuperadmins'
    assert ldapsuperadmins.description == 'Super admins of the LDAP directory'
    assert ldapsuperadmins.company_id == 'ee'
    assert ldapsuperadmins.member_usernames == ['brenard']
    assert ldapsuperadmins.member_dns == ['uid=brenard,ou=people,o=eesyldap']
    assert ldapsuperadmins.owner_dn == 'uid=brenard,ou=people,o=eesyldap'


def test_group_relations_access(ldapsuperadmins, brenard, ee):
    """ Test access to group relations """

    # Members
    members = ldapsuperadmins.members
    assert len(members) == 1
    assert brenard.username in members
    assert members[brenard.username] == brenard

    # Owner
    assert ldapsuperadmins.owner == brenard

    # User owned group
    assert ldapsuperadmins.name in brenard.owned_groups

    # Company
    assert ldapsuperadmins.company == ee

    # Services
    services = ldapsuperadmins.services
    assert isinstance(services, dict)
    assert not services


def test_group_add_member(ldapsuperadmins, jdoe):
    """ Test to add a user to group as member """
    members = ldapsuperadmins.members.copy()
    members[jdoe.dn] = jdoe
    ldapsuperadmins.members = members

    assert jdoe.username in ldapsuperadmins.member_usernames
    assert jdoe.dn in ldapsuperadmins.member_dns


def test_group_remove_member(ldapsuperadmins, brenard):
    """ Test to remove a group's member """
    members = ldapsuperadmins.members.copy()
    del members[brenard.username]
    ldapsuperadmins.members = members

    assert brenard.username not in ldapsuperadmins.member_usernames
    assert brenard.dn not in ldapsuperadmins.member_dns


def test_group_remove_all_members(ldapsuperadmins):
    """ Test to remove all group's members """
    ldapsuperadmins.members = dict()

    assert not ldapsuperadmins.member_usernames
    assert not ldapsuperadmins.member_dns


def test_group_delete_member(ldapsuperadmins, jdoe):
    """ Test to delete a group's member """
    members = ldapsuperadmins.members.copy()
    members[jdoe.dn] = jdoe
    ldapsuperadmins.members = members

    jdoe.delete()
    ldapsuperadmins.refresh()

    assert jdoe.username not in ldapsuperadmins.member_usernames
    assert jdoe.dn not in ldapsuperadmins.member_dns



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP connection """

import logging
import time
import threading

import ldap
import ldap.modlist
from ldap import INVALID_CREDENTIALS        # pylint: disable=no-name-in-module
from ldap import SIZELIMIT_EXCEEDED         # pylint: disable=no-name-in-module

from eesyldap.exceptions import LDAPConnectionError
from eesyldap.exceptions import LDAPInvalidCredentials
from eesyldap.exceptions import LDAPSearchHardLimitExceeded
from eesyldap.exceptions import LDAPSearchLimitExceeded

log = logging.getLogger(__name__)


class LDAPConnection:
    """
    LDAP connection

    This class handle all LDAP connection stuff and locking problems.
    It's also implement helpers above python LDAP method to interact with LDAP server.

    param: config:  LDAP connection configuration as dict :

        - uri:              LDAP connection URI (required)
        - tls:              Enable TLS (optional, default: False)
        - base_dn:          Base DN (required)
        - bind_dn:          Bind DN (optional, default: annonymous connection)
        - bind_credential:  Bind credential (optional, default: annonymous connection)
        - timeout:          Operation and network connection timeout in seconds (optional, default: 2 seconds)
    """

    def __init__(self, config):
        self.config = config
        self.config['timeout'] = float(config['timeout']) if config.get('timeout') is not None else 2.0
        self.__lock = threading.Lock()
        self.con = None

    def __enter__(self):
        """ Acquire lock and bind on LDAP server """
        self.__lock.acquire()
        return self.bind(
            self.config.get('bind_dn'),
            self.config.get('bind_credential')
        )

    def __exit__(self, exit_type, value, traceback):
        """ Unbind on LDAP server and release lock """
        self.unbind()
        self.__lock.release()

    def bind(self, dn=None, credential=None, force=False):
        """ Initialize LDAP connection """

        if self.con is not None and not force:
            return True

        try:
            log.debug('Connect to LDAP server "%s"', self.config['uri'])
            self.con = ldap.initialize(self.config['uri'])

            log.debug('Set LDAP connection timeouts to %s second(s)', self.config['timeout'])
            self.con.set_option(ldap.OPT_NETWORK_TIMEOUT, self.config['timeout'])  # pylint: disable=no-member
            self.con.set_option(ldap.OPT_TIMEOUT, self.config['timeout'])          # pylint: disable=no-member
            self.con.set_option(ldap.OPT_PROTOCOL_VERSION, ldap.VERSION3)          # pylint: disable=no-member

            if self.config.get('tls', 'false').lower() == 'true':
                self.con.start_tls_s()

            if dn and credential:
                log.debug('Bind on LDAP server "%s" with dn "%s"', self.config['uri'], dn)
                self.con.simple_bind_s(dn, credential)
            else:
                log.debug('No DN and password configured. Use annonymous connection.')

            return True
        except INVALID_CREDENTIALS:  # pylint: disable=no-member
            log.error('[LDAP] Authentication failed (%s)', dn)

            # Close the connection
            self.unbind()

            raise LDAPInvalidCredentials(self.config['uri'], dn)

        except Exception as err:
            log.error('Failed to connect to remote server', exc_info=1)

            # Close connection
            self.unbind()

            # Raise error
            raise LDAPConnectionError(self.config['uri'], str(err))

        return True

    def unbind(self):
        """ Unbind on LDAP connection """
        if self.con is None:
            return False

        try:
            self.con.unbind()

        except Exception:
            log.warning('Failed to unbind connection')
            return False

        finally:
            self.con = None

        return True

    @staticmethod
    def _get_scope(scope=None):
        """ Convert scope parameter value for python-ldap """
        if scope == 'sub' or scope is None:
            return ldap.SCOPE_SUBTREE   # pylint: disable=no-member
        if scope == 'one':
            return ldap.SCOPE_ONELEVEL  # pylint: disable=no-member
        if scope == 'base':
            return ldap.SCOPE_BASE      # pylint: disable=no-member
        raise Exception('Unknwon LDAP search scope %s' % scope)

    def search(self, base=None, filterstr='(objectClass=*)', scope=None, limit=None, hardlimit=None, **kwargs):
        """ LDAP search helper """
        if base is None:
            base = self.config.get('base_dn')
        if scope is None:
            scope = 'sub'
        log.debug('Run LDAP search with filter "%s" on base "%s" (scope = %s)', filterstr, base, scope)
        scope = self._get_scope(scope)
        sizelimit = 0 if hardlimit is None else int(hardlimit)
        try:
            start_time = time.time()
            result = self.con.search_ext_s(base, scope, filterstr=filterstr, sizelimit=sizelimit, **kwargs)
            log.debug('LDAP search result retreived in %.3f second(s)', (time.time() - start_time))
            # Ensure hard limit is handled
            if hardlimit and len(result) > hardlimit:
                raise LDAPSearchHardLimitExceeded(hardlimit)

            # Filter result to keep only items with DN (ignore references)
            result = [obj for obj in result if obj[0]]

            # Handle limit
            if limit and len(result) > limit:
                raise LDAPSearchLimitExceeded(limit)

            return result
        except SIZELIMIT_EXCEEDED:
            raise LDAPSearchHardLimitExceeded(sizelimit)

    def add(self, dn, modlist):
        """ LDAP add helper """
        return self.con.add_s(dn, modlist)

    def modify(self, dn, modlist, **kwargs):
        """ LDAP modify helper """
        return self.con.modify_s(dn, modlist, **kwargs)

    def rename(self, dn, new_rdn, **kwargs):
        """ LDAP modify helper """
        new_rdn = new_rdn.decode('utf-8') if isinstance(new_rdn, bytes) else new_rdn
        return self.con.rename_s(dn, new_rdn, **kwargs)

    def delete(self, dn):
        """ LDAP delete helper """
        return self.con.delete_s(dn)
